unit ForEach1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls,
  OtlCommon,
  OtlCollections,
  OtlParallel;

type
  TfrmForEach = class(TForm)
    lbLog: TListBox;
    btnSeqPrimes: TButton;
    btnParallelPrimes: TButton;
    btnParallelPrimesInto: TButton;
    btnOrderedPrimes: TButton;
    btnCountPrimesSeq: TButton;
    btnCountPrimesPara: TButton;
    btnCountPrimesAgg: TButton;
    btnIterateOverContainer: TButton;
    btnIterateOverContainer2: TButton;
    btnFrequencySeq: TButton;
    btnFreqPara: TButton;
    procedure FormDestroy(Sender: TObject);
    procedure btnCountPrimesAggClick(Sender: TObject);
    procedure btnCountPrimesParaClick(Sender: TObject);
    procedure btnCountPrimesSeqClick(Sender: TObject);
    procedure btnFreqParaClick(Sender: TObject);
    procedure btnFrequencySeqClick(Sender: TObject);
    procedure btnIterateOverContainer2Click(Sender: TObject);
    procedure btnIterateOverContainerClick(Sender: TObject);
    procedure btnOrderedPrimesClick(Sender: TObject);
    procedure btnParallelPrimesIntoClick(Sender: TObject);
    procedure btnParallelPrimesClick(Sender: TObject);
    procedure btnSeqPrimesClick(Sender: TObject);
  private
    FStringData: TStringList;
    procedure VerifyResult;
    procedure CreateSampleData;
    function CountLetters(const s: string; letter: char): integer;
  public
  end;

var
  frmForEach: TfrmForEach;

implementation

uses
  GpLists,
  Character,
  DSiWin32;

{$R *.dfm}

const
  CMaxCountTest = 1000000;
  CMaxTest = 20000;
  CMaxData = 1000000;

function IsPrime(i: integer): boolean;
var
  j: integer;
begin
  Result := false;
  if i <= 1 then
    Exit;
  for j := 2 to Round(Sqrt(i)) do
    if (i mod j) = 0 then
      Exit;
  Result := true;
end;

{ TfrmForEach }

procedure TfrmForEach.VerifyResult;
var
  error : boolean;
  iItem : integer;
  order : string;
  primes: TGpIntegerList;
  value : integer;
  value2: integer;
begin
  if lbLog.Items.Count <= 1 then
    lbLog.Items.Add('Error, empty result list!')
  else begin
    error := false;
    order := 'ordered';
    value := StrToInt(lbLog.Items[0]);
    primes := TGpIntegerList.Create;
    try
      primes.Add(value);
      for iItem := 1 to lbLog.Items.Count - 1 do begin
        value2 := StrToInt(lbLog.Items[iItem]);
        if value2 <= value then
          order := 'unordered';
        primes.Add(value2);
        value := value2;
      end;
      primes.Sort;
      for iItem := 1 to CMaxTest do begin
        if IsPrime(iItem) then begin
          if not primes.Contains(iItem) then begin
            error := true;
            break; //for
          end
        end
        else if primes.Contains(iItem) then begin
          error := true;
          break; //for
        end;
      end; //for
      if error then
        lbLog.Items.Add('ERROR, list is ' + order)
      else
        lbLog.Items.Add('OK, list is ' + order);
    finally FreeAndNil(primes); end;
  end;
  lbLog.ItemIndex := lbLog.Items.Count - 1;
end;

procedure TfrmForEach.btnCountPrimesSeqClick(Sender: TObject);
var
  numPrimes: integer;
  prime    : integer;
  start    : int64;
begin
  start := DSiTimeGetTime64;

  numPrimes := 0;
  for prime := 1 to CMaxCountTest do
    if IsPrime(prime) then
      Inc(numPrimes);

  start := DSiElapsedTime64(start);
  lbLog.Items.Add(Format('%d primes from 1 to %d; %d ms', [numPrimes, CMaxCountTest, start]));
end;

procedure TfrmForEach.btnCountPrimesParaClick(Sender: TObject);
var
  numPrimes: integer;
  start    : int64;
begin
  start := DSiTimeGetTime64;
  numPrimes := 0;
  Parallel.ForEach(1, CMaxCountTest).Execute(
    procedure (const prime: integer)
    begin
      if IsPrime(prime) then
        InterlockedIncrement(numPrimes);
    end);

  start := DSiElapsedTime64(start);
  lbLog.Items.Add(Format('%d primes from 1 to %d; %d ms', [numPrimes, CMaxCountTest, start]));
end;

procedure TfrmForEach.btnCountPrimesAggClick(Sender: TObject);
var
  numPrimes: integer;
  start    : int64;
begin
  start := DSiTimeGetTime64;
  numPrimes :=
    Parallel.ForEach(1, CMaxCountTest)
    .Aggregate(0,
      procedure (var aggregate: TOmniValue; const value: TOmniValue)
      begin
        aggregate := aggregate.AsInt64 + value.AsInt64;
      end)
    .Execute(
      procedure (const value: integer; var result: TOmniValue)
      begin
        if IsPrime(value) then
          Result := 1;
      end
    );

  start := DSiElapsedTime64(start);
  lbLog.Items.Add(Format('%d primes from 1 to %d; %d ms', [numPrimes, CMaxCountTest, start]));
end;

procedure TfrmForEach.CreateSampleData;

  function GenerateRandomString: string;
  var
    i: integer;
  begin
    SetLength(Result, Random(1024));
    for i := 1 to Length(Result) do
      Result[i] := Chr(32 + Random(96));
  end;

var
  i: integer;

begin
  if assigned(FStringData) then
    Exit;

  lbLog.Items.Add('Generating sample data...'); lbLog.Update;
  RandSeed := 42;
  FStringData := TStringList.Create;
  for i := 1 to CMaxData do
    FStringData.Add(GenerateRandomString);
  lbLog.Items.Add('Data generated'); lbLog.Update;
end;

function TfrmForEach.CountLetters(const s: string; letter: char): integer;
var
  ch: char;
begin
  Result := 0;
  for ch in s do
    if letter = ch then
      Inc(Result);
end;

procedure TfrmForEach.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FStringData);
end;

procedure TfrmForEach.btnFrequencySeqClick(Sender: TObject);
var
  freq: array of integer;
  i   : integer;
  start: int64;
begin
  CreateSampleData;
  SetLength(freq, FStringData.Count);
  start := DSiTimeGetTime64;

  for i := 0 to FStringData.Count - 1 do
    freq[i] := CountLetters(FStringData[i], 'z');

  start := DSiElapsedTime64(start);
  lbLog.Items.Add(Format('%d ms; %d, %d, %d, %d, %d', [start, freq[0], freq[1], freq[2], freq[3], freq[4]]));
end;

procedure TfrmForEach.btnFreqParaClick(Sender: TObject);
var
  freq: array of integer;
  start: int64;
begin
  CreateSampleData;
  SetLength(freq, FStringData.Count);
  start := DSiTimeGetTime64;

  Parallel.ForEach(0, FStringData.Count - 1).Execute(
    procedure (const i: integer)
    begin
      freq[i] := CountLetters(FStringData[i], 'z');
    end);

  start := DSiElapsedTime64(start);
  lbLog.Items.Add(Format('%d ms; %d, %d, %d, %d, %d', [start, freq[0], freq[1], freq[2], freq[3], freq[4]]));
end;

procedure TfrmForEach.btnSeqPrimesClick(Sender: TObject);
var
  prime: integer;
begin
  lbLog.Clear;
  for prime := 1 to CMaxTest do
    if IsPrime(prime) then
      lbLog.Items.Add(IntToStr(prime));
  VerifyResult;
end;

procedure TfrmForEach.btnParallelPrimesClick(Sender: TObject);
var
  forEach   : IOmniParallelLoop<integer>;
  prime     : TOmniValue;
  primeQueue: IOmniBlockingCollection;
begin
  lbLog.Clear;
  primeQueue := TOmniBlockingCollection.Create;

  forEach := Parallel.ForEach(1, CMaxTest);
  forEach.NoWait;
  forEach.OnStop(
    procedure
    begin
      primeQueue.CompleteAdding;
    end);
  forEach.Execute(
    procedure (const value: integer)
    begin
      if IsPrime(value) then begin
        primeQueue.Add(value);
      end;
    end);

  for prime in primeQueue do begin
    lbLog.Items.Add(IntToStr(prime));
    lbLog.Update;
  end;
  VerifyResult;
end;

procedure TfrmForEach.btnParallelPrimesIntoClick(Sender: TObject);
var
  prime     : TOmniValue;
  primeQueue: IOmniBlockingCollection;
 begin
  lbLog.Clear;
  primeQueue := TOmniBlockingCollection.Create;

  Parallel.ForEach(1, CMaxTest).NoWait
    .Into(primeQueue)
    .Execute(
      procedure (const value: integer; var res: TOmniValue)
      begin
        if IsPrime(value) then
          res := value;
      end);

  for prime in primeQueue do begin
    lbLog.Items.Add(IntToStr(prime));
    lbLog.Update;
  end;
  VerifyResult;
end;

procedure TfrmForEach.btnOrderedPrimesClick(Sender: TObject);
var
  forEach   : IOmniParallelLoop<integer>;
  prime     : TOmniValue;
  primeQueue: IOmniBlockingCollection;
begin
  lbLog.Clear;
  primeQueue := TOmniBlockingCollection.Create;

  forEach := Parallel.ForEach(1, CMaxTest);
  forEach.PreserveOrder;
  forEach.NoWait;
  forEach.Into(primeQueue)
    .Execute(
      procedure (const value: integer; var res: TOmniValue)
      begin
        if IsPrime(value) then
          res := value;
      end);

  for prime in primeQueue do begin
    lbLog.Items.Add(IntToStr(prime));
    lbLog.Update;
  end;
  VerifyResult;
end;

procedure TfrmForEach.btnIterateOverContainerClick(Sender: TObject);
var
  i         : integer;
  numNumbers: integer;
  s         : TStringList;
begin
  s := TStringList.Create;
  try
    for i := 0 to 65535 do
      s.Add(Chr(i));

    numNumbers := 0;
    Parallel.ForEach(s).Execute(
      procedure (const value: TOmniValue)
      begin
        if TCharacter.IsNumber(value.AsString[1]) then
          InterlockedIncrement(numNumbers);
      end
    );

    lbLog.Items.Add(Format('%d numbers in Unicode basic plane', [numNumbers]));
  finally s.Free; end;
end;

procedure TfrmForEach.btnIterateOverContainer2Click(Sender: TObject);
var
  i         : integer;
  numNumbers: integer;
  s         : TStringList;
begin
  s := TStringList.Create;
  try
    for i := 0 to 65535 do
      s.Add(Chr(i));

    numNumbers := 0;
    Parallel.ForEach<string>(s).Execute(
      procedure (const value: string)
      begin
        if TCharacter.IsNumber(value[1]) then
          InterlockedIncrement(numNumbers);
      end
    );

    lbLog.Items.Add(Format('%d numbers in Unicode basic plane', [numNumbers]));
  finally s.Free; end;
end;

end.
