unit Pipeline1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls,
  OtlTask,
  OtlTaskControl,
  OtlCommon,
  OtlComm,
  OtlCollections,
  OtlParallel;

const
  WM_STOPPED = WM_USER + 1;

type
  TfrmWebSpider = class(TForm)
    inpURL: TEdit;
    btnStart: TButton;
    lbLog: TListBox;
    procedure btnStartClick(Sender: TObject);
  private
    FSpider: IOmniPipeline;
    FTargetHost: string;
    FURLCount: IOmniCounter;
    procedure btnStopClick(Sender: TObject);
    procedure UniqueFilter(const input, output: IOmniBlockingCollection);
    procedure HttpGet(const input: TOmniValue; var output: TOmniValue);
    procedure StoreToDB(const input, output: IOmniBlockingCollection; const task: IOmniTask);
    procedure ShowStored(const task: IOmniTaskControl; const msg: TOmniMessage);
    procedure ExtractLinks(const input: TOmniValue; var output: TOmniValue);
    procedure WMStopped(var msg: TMessage); message WM_STOPPED;
  public
  end;

var
  frmWebSpider: TfrmWebSpider;

implementation

uses
  StrUtils,
  MMSHelpers,
  HTMLParser;

{$R *.dfm}

procedure TfrmWebSpider.UniqueFilter(const input, output: IOmniBlockingCollection);
var
  uniqueUrls: TStringList;
  url       : string;
begin
  uniqueUrls := TStringList.Create;
  try
    uniqueUrls.Sorted := true;
    for url in input do begin
      if uniqueUrls.IndexOf(url) < 0 then begin
        uniqueUrls.Add(url);
        output.TryAdd(url);
      end
      else if FURLCount.Decrement = 0 then
        FSpider.Input.CompleteAdding;
    end;
  finally FreeAndNil(uniqueUrls); end;
end;

procedure TfrmWebSpider.HttpGet(const input: TOmniValue; var output: TOmniValue);
var
  pageContents: string;
begin
  if MMSHelpers.HttpGet(FTargetHost, 80, input, pageContents, '') then
    output := TOmniValue.CreateNamed(['URL', input.AsString, 'Page', pageContents])
  else if FURLCount.Decrement = 0 then
    FSpider.Input.CompleteAdding;
end;

procedure TfrmWebSpider.StoreToDB(const input, output: IOmniBlockingCollection;
  const task: IOmniTask);
var
  page: TOmniValue;
begin
  // open database
  for page in input do begin
    // store page['URL'] and page['Page'] into the database
    task.Comm.Send(0, page['URL']);
    output.TryAdd(page['Page']);
  end;
  // close database
end;

procedure TfrmWebSpider.ExtractLinks(const input: TOmniValue; var output: TOmniValue);
var
  parser: THTMLParser;
  url: string;
begin
  parser := THTMLParser.Create(nil);
  try
    parser.HTML.Add(input);
    parser.Execute;
    for url in parser.Parsed.Hyperlinks do
      if not StartsText('http', url) then begin
        FURLCount.Increment;
        FSpider.Input.TryAdd(url);
      end;
    if FURLCount.Decrement = 0 then
      FSpider.Input.CompleteAdding;
  finally FreeAndNil(parser); end;
end;

procedure TfrmWebSpider.ShowStored(const task: IOmniTaskControl; const msg: TOmniMessage);
begin
  lbLog.Items.Add(msg.MsgData);
end;

procedure TfrmWebSpider.btnStartClick(Sender: TObject);
begin
  btnStart.Caption := 'Stop';
  btnStart.OnClick := btnStopClick;
  lbLog.Clear;
  FTargetHost := inpURL.Text;

  FSpider := Parallel.Pipeline
    .Stage(UniqueFilter)
    .Stage(HttpGet).NumTasks(Environment.Process.Affinity.Count)
    .Stage(StoreToDB, Parallel.TaskConfig.OnMessage(ShowStored)).NumTasks(1)
    .Stage(ExtractLinks).NumTasks(2)
    .OnStop(
      procedure
      begin
        PostMessage(frmWebSpider.Handle, WM_STOPPED, 0, 0);
      end)
    .Run;

  FURLCount := CreateCounter(1);
  FSpider.Input.Add('/');
end;

procedure TfrmWebSpider.btnStopClick(Sender: TObject);
begin
  btnStart.Caption := 'Start';
  btnStart.OnClick := btnStartClick;

  FSpider.Input.CompleteAdding;
  FSpider.WaitFor(INFINITE);
  FSpider := nil;
end;

procedure TfrmWebSpider.WMStopped(var msg: TMessage);
begin
  btnStopClick(nil);
end;

end.
